# Filename:        helpers.R           
# Date:            2017-03-08
# Author:          Kristian Sandaker (kristian.sandaker@outlook.com)
# Purpose:         Functions for e.g. loading files
# Data Written:    
# Packages Used:   


importDataFrame <- function(dataFrameAddress, startDate, endDate,
                            firstColDateCol=T, dateFormat='%d/%m/%Y',
                            excludeVector=c('norwegianexports_norgessjomatrad', 
                                            'forward.eur.Xm_fishpool'),
                            convertToReturns = '',
                            timeWindow=1) {
  df <- read.csv(dataFrameAddress, stringsAsFactors=F)
  startIndex <- 1
  if (firstColDateCol) {
    df[,1] <- as.Date(df[,1], format=dateFormat)
    df <- df[df[,1] >= startDate & df[,1] <= endDate,]
    df <- df[,!(colnames(df) %in% excludeVector)]
    colnames(df)[1] <- 'date'
    startIndex <- 2
  }
  for (i in startIndex:dim(df)[2]) {
    df[,i] <- as.numeric(df[,i]) # Convert to numerics
  }
  # Now do eventual returns conversion
  if (firstColDateCol && convertToReturns != "") {
    relevantDates <- df$date
    n_col <- dim(df)[2]
    n_row <- dim(df)[1]
    
    if (convertToReturns == 'log') {
      df <- log(df[(1 + timeWindow):n_row, 2:n_col]) - log(df[1:(n_row - timeWindow), 2:n_col])
    } else if (convertToReturns == 'arithmetic') {
      df <- (df[(1 + timeWindow):n_row, 2:n_col] - df[1:(n_row - timeWindow), 2:n_col])/
        df[1:(n_row - timeWindow), 2:n_col]
    }
    
    df[df == 'NaN'] <- NA
    df[sapply(df, is.infinite)] <- NA
    df[is.na(df)] <- 0
    df <- data.frame(date=relevantDates[(1 + timeWindow):n_row], df)
  }
  return(df)
}

importLagStructureDataFrame <- function(lagDataFrameAddress,
                                        excludeVector=c('norwegianexports_norgessjomatrad', 
                                                        'forward.eur.Xm_fishpool')) {
  df <- read.csv(lagDataFrameAddress, header=F, stringsAsFactors=F)
  df <- df[!(df[,1] %in% excludeVector),]
  df[,1] <- gsub("-", ".", df[,1])
  return(df)
}

importTimeWindowsDataFrame <- function(timeWindowsDataFrameAddress) {
  df <- read.csv(timeWindowsDataFrameAddress, header=F, stringsAsFactors=F)
  df[,1] <- gsub("-", ".", df[,1])
  return(df)
}

importGroupStructureDataFrame <- function(groupStructureDataFrameAddress,
                                          constraints=F) {
  df <- read.csv(groupStructureDataFrameAddress, header=constraints, stringsAsFactors=F)
  df[,1] <- gsub("-", ".", df[,1])
  return(df)
}

importFullLaggedDataFrame <- function(dataFrameAddress, lagDataFrameAddress,
                                      useLagDataFrame=T, maxLag=24,
                                      startDate, endDate, firstColDateCol=T, 
                                      dateFormat='%d/%m/%Y',
                                      excludeVector=c('norwegianexports_norgessjomatrad', 
                                                      'forward.eur.Xm_fishpool'),
                                      convertToReturns='', forecastHorizon=12,
                                      useWindowDataFrame=F, windowDataFrameAddress='',
                                      useTimeWindowsDataFrame=F,
                                      timeWindowsDataFrameAddress='',
                                      allowedTimeWindows=c(1,12)) {
  
  df_series <- importDataFrame(dataFrameAddress=dataFrameAddress, startDate=minDate,
                            endDate=maxDate, firstColDateCol=T, dateFormat='%d/%m/%Y',
                            excludeVector=excludeVector, convertToReturns='log', timeWindow=1)
  
  if (useLagDataFrame) {
    df_lagStructure <- importLagStructureDataFrame(lagDataFrameAddress,
                                                   excludeVector)
  } else {
    n_series <- dim(df_series)[2] - 1
    matr <- rep(1:maxLag, each=(n_series))
    dim(matr) <- c(n_series, maxLag)
    df_lagStructure <- data.frame(colnames(df_series)[2:(1 + n_series)],matr)
  }
  print(df_lagStructure)
  
  if (useTimeWindowsDataFrame) {
    df_timeWindows <- importTimeWindowsDataFrame(timeWindowsDataFrameAddress)
    maxTimeWindow <- max(df_timeWindows[,2:dim(df_timeWindows)[2]])
  } else {
    maxTimeWindow <- max(allowedTimeWindows)
  }
  
  n_lagStructure <- dim(df_lagStructure)[2]
  T_min <- max(df_lagStructure[,2:n_lagStructure]) + 2 + max(maxTimeWindow) # Relative time delay
  T_max <- dim(df_series)[1]
  validSeriesNames <- colnames(df_series)[2:dim(df_series)[2]]
  relevantDates <- df_series[,1][(T_min):dim(df_series)[1]]
  relevantSeriesNames <- as.character(df_lagStructure[,1])
  print(relevantSeriesNames)
  out_df <- data.frame(date=relevantDates)
  
  # Loop through each data series and lag acoording to the respective lagStructure
  for (i in 1:length(relevantSeriesNames)) {
    relevantSerieName <- relevantSeriesNames[i]
    print(paste("Working on series: ", relevantSerieName, sep = ""))
    if (relevantSerieName %in% validSeriesNames) {
      
      maxLag <- max(df_lagStructure[which(df_lagStructure[,1] == relevantSerieName),2:dim(df_lagStructure)[2]])
      
      for (lag in 1:maxLag) {
        # Import for forecast
        if (relevantSerieName == 'salmonprice_usdkg_nasdaq') {
          for (k in 1:forecastHorizon) {
            y_lag <- k - 1
            laggedSeries <- df_series[relevantSerieName][(T_min+y_lag):(T_max+y_lag),1] # Eventual pad with NAs
            if (k > 1) {
              for (l in 2:k) {
                laggedSeries <- laggedSeries + df_series[relevantSerieName][(T_min+y_lag-(l-1)):(T_max+y_lag-(l-1)),1]
              }
            }
            df_colName <- paste(relevantSerieName, '_fc.', sprintf('%02d', y_lag), sep='')
            out_df[df_colName] <- laggedSeries
          }
        }
        # Import for explanatory variables...
        if ((!useWindowDataFrame) && (lag > 0)) {
          if (useTimeWindowsDataFrame) {
            allowedTimeWindows <- df_timeWindows[which(df_timeWindows[,1]==relevantSerieName),][1,2:dim(df_timeWindows)[2]]
            if (length(allowedTimeWindows) == 0) { # Quick and dirty exception handling
              break
            }
          }
          for (k in 1:length(allowedTimeWindows)) {
            window = as.numeric(allowedTimeWindows[k])
            if (window == -1) {
              break
            }
            df_laggedWindow <- importDataFrame(dataFrameAddress, startDate, endDate, firstColDateCol, dateFormat,
                                               excludeVector, convertToReturns, timeWindow=window)
            df_colName <- paste(relevantSerieName, '_lag.', sprintf('%02d', lag), '_delta.', sprintf('%02d', window), sep='')
            
            df_inFrame <- df_laggedWindow[relevantSerieName][(T_min-lag-window+1):(T_max-lag-window+1),]
            out_df[df_colName] <- df_inFrame
            
            printStr <- sprintf('%10s %20s %8s %4d %12s %4d %12s %8d', "Importing", relevantSerieName,
                                "Lag", lag, "Time window", window, "Length", length(df_inFrame))
            print(printStr)
            
            #print(paste("Importing ", relevantSerieName, " at lag ", lag, ", time window ", k, 
            #            " @ length ", length(df_inFrame), sep=""))
          }
        }
      }
    }
  }
  return(out_df)
}








