# Filename:        latex_output.R           
# Date:            2017-03-08
# Author:          Kristian Sandaker (kristian.sandaker@outlook.com)
# Purpose:         Functions for LaTeX tables and plotting
# Data Written:    data/dataset.csv
# Packages Used:   np, moments, stargazer, tseries, urca, zoo

library(np)
library(moments)
library(stargazer)
library(tseries)
library(urca)
library(zoo)


source('helpers.R')
source('quantile_regression_model.R')


createDescriptiveStatsTable <- function(all = TRUE, dataFrameAddress, outFileAddress,
                                        outputNamesDataFrameAddress, lagDataFrameAddress,
                                        outFileDir = '/Users/Kristian/Kristian/Dropbox/Apps/ShareLaTeX/Masteroppgave/tables/',
                                        useStargazer = TRUE, validDates = "", 
                                        startDate_descStats = as.Date('2005-01-01'),
                                        endDate_descStats = as.Date('2017-01-31'), 
                                        removeDates = FALSE, addResizeBox = TRUE, resizeBoxFactor=1.7,
                                        ljungBoxRho = 6) {

  # Some introductory parameters to the function
  datesColumn <- TRUE
  
  # Now run the import
  df_logReturns <- importDataFrame(dataFrameAddress=dataFrameAddress, 
                                   startDate=as.Date('2005-01-01'), endDate=as.Date('2017-01-31'),
                                   convertToReturns='log')
  df_arithmeticReturns <- importDataFrame(dataFrameAddress=dataFrameAddress, 
                                          startDate=as.Date('2005-01-01'), endDate=as.Date('2017-01-31'),
                                          convertToReturns='arithmetic')
  
  df_logReturns <- df_logReturns[which(df_logReturns[,1] >= startDate_descStats & df_logReturns[,1] <= endDate_descStats),]
  df_arithmeticReturns <- df_arithmeticReturns[which(df_arithmeticReturns[,1] >= startDate_descStats & df_arithmeticReturns[,1] <= endDate_descStats),]
  
  df_outputNames <- importGroupStructureDataFrame(outputNamesDataFrameAddress, constraints=F)
  df_lagStructure <- importLagStructureDataFrame(lagDataFrameAddress,
                                                 excludeVector=c())
  
  # Select particular column names
  if (!all) {
    relevantSeries <- c("date")
    relevantSeriesNames <- c()
    for (i in 2:dim(df_logReturns)[2]) {
      seriesName <- colnames(df_logReturns)[i]
      df_lagStructure_index <- which(df_lagStructure[,1]==seriesName)
      if (sum(df_lagStructure[df_lagStructure_index,2:dim(df_lagStructure)[2]] > 0) > 0) {
        relevantSeries <- c(relevantSeries, df_lagStructure[df_lagStructure_index,1])
        relevantSeriesNames <- c(relevantSeriesNames, df_outputNames[df_lagStructure_index,2])
      }
    }
    df_logReturns <- df_logReturns[relevantSeries]
    df_arithmeticReturns <- df_arithmeticReturns[relevantSeries]
  }
  
  # Date correction
  # df_timeSeries <- df_timeSeries[which(df_timeSeries[,1] >= startDate & df_timeSeries[,1] <= endDate),]
  # df_logReturns <- df_logReturns[which(df_logReturns[,1] >= startDate & df_logReturns[,1] <= endDate),]
  
  # Only select certain dates
  # if (length(validDates) > 1) {
  #   df_logReturns <- df_logReturns[df_logReturns$date %in% validDates,]
  #   df_arithmeticReturns <- df_arithmeticReturns[df_arithmeticReturns$date %in% validDates,]
  # } else {
  #   df_logReturns <- df_logReturns[which((df_logReturns$date >= minDate) & (df_logReturns$date <= maxDate)),]
  #   df_arithmeticReturns <- df_arithmeticReturns[which((df_arithmeticReturns$date >= minDate) & (df_arithmeticReturns$date <= maxDate)),]
  # }
  
  # Put all series in a single list
  listNames <- c("descstats-logreturns", "descstats-arithmeticreturns")
  mainNames <- c("Descriptive statistics of the log returns of the data series",
                 "Descriptive statistics of the arithmetic returns of the data series") # \\\hspace{\textwidth}This is the second line
  mainNames <- paste(mainNames, ", time period: ", format(startDate_descStats, "%b. %Y"), " -- ", format(endDate_descStats, "%b. %Y"), sep="")
  operatorNames <- c("ln", "%")
  descStatList <- list(df_logReturns, df_arithmeticReturns)
  
  ### DESCRIPTIVE STATISTICS -- LOG RETURNS ###
  if (!removeDates) {
    descStatTableColNames <- c("names", "nobs", "mindate", "maxdate", "mean", "stddev",
                               "minobs", "maxobs", "skew", "exkurt", "corr", "JB", "DF", "ADF", "Ljung-Box")
  } else {
    descStatTableColNames <- c("names", "nobs", "mean", "stddev",
                               "minobs", "maxobs", "skew", "exkurt", "corr", "JB", "DF", "ADF", "Ljung-Box")
    latexTableColNames <- c("Series name", "N", "Mean", "Std. dev.", "Min", "Max", "Skew",
                            "Ex. kurt.", "NQSALMON corr.", "JB", "DF", "ADF", "Ljung-Box")
  }
  latexTableColNames <- c("Series name", "N", "Mean", "Std. dev.", "Min", "Max", "Skew",
                          "Ex. kurt.", "NQSALMON corr.", "JB", "DF", "ADF", "Ljung-Box")
  latexTableUnits <- list()
  latexTableUnits[[1]] <- c(" ", "\\#", "{ }", "{ }", "{ }", "{ }", "{ }", "{ }", "{ }", "{ }", "{ }", "{ }", paste("{$\\rho=", ljungBoxRho, "$}", sep = ""))
  latexTableUnits[[2]] <- c(" ", "\\#", "{\\%}", "{\\%}", "{\\%}", "{\\%}", "{ }", "{ }", "{ }", "{ }", "{ }", "{ }", paste("{$\\rho=", ljungBoxRho, "$}", sep = ""))
  
  
  for (d in 1:length(descStatList)) {
    
    # jb <- jarque.bera.test((blm$residuals)) #1. JB test on residuals
    # lb <- Box.test(blm$residuals, lag = maxLags, type = "Ljung-Box") #3. Ljung-Box serial corr. on residuals
    # resultsFrame[j, "JB"] <- paste("$", round(jb$statistic,2), getFstars(jb$p.value), "$", sep = "")
    # resultsFrame[j, "Ljung-Box"] <- paste("$", round(lb$statistic,2), getFstars(lb$p.value), "$", sep = "")
    
    df_descStats <- descStatList[[d]]
    
    descriptive_stats <- data.frame(matrix("", dim(df_logReturns)[2] - 1, length(descStatTableColNames)), stringsAsFactors=F)
    colnames(descriptive_stats) <- descStatTableColNames
    
    descriptive_stats[,1] <- paste("$\\Delta\\", operatorNames[d], "$[", relevantSeriesNames, "]", sep="") # Set rows to first column
    
    for (i in 2:dim(df_descStats)[2]) {
      
      # Note that all series have to be pushed row-wise backward due to the first decsStats column being a Date
      descriptive_stats[i-1,2] <- length(which(!is.na(df_descStats[,i]))) # N observations
      
      # First run some of the tests
      jb_test <- jarque.bera.test(df_descStats[which(!is.na(df_descStats[,i])),i]) # Addresses by $statistic
      df_test <- ur.df(df_descStats[which(!is.na(df_descStats[,i])),i], type = "none", lags = 0) # Addressed by @teststat[1]
      adf_test <- ur.df(df_descStats[which(!is.na(df_descStats[,i])),i], type = "none", lags = 1) # Addressed by @teststat[1]
      lb_test <- Box.test(df_descStats[which(!is.na(df_descStats[,i])),i], lag = ljungBoxRho, type = "Ljung-Box") #3. Ljung-Box serial corr. on residuals
      
      # Number penalty
      
      options(scipen=100, digits=3)
      
      if (!removeDates) {
        descriptive_stats[i-1,3] <- paste("", df_descStats[min(which(!is.na(df_descStats[,i]))),1], sep = "") # Min dates as string
        descriptive_stats[i-1,4] <- paste("", df_descStats[max(which(!is.na(df_descStats[,i]))), 1], sep = "") # Max dates as string
        descriptive_stats[i-1,5] <- sprintf("%0.8f", mean(df_descStats[which(!is.na(df_descStats[,i])),i])*(1+99*(d==2))) # Mean sprintf("%4.2e", mean(dfLogRets[,2]))
        descriptive_stats[i-1,6] <- sprintf("%0.8f", sd(df_descStats[which(!is.na(df_descStats[,i])),i])*(1+99*(d==2))) # Std. dev.
        descriptive_stats[i-1,7] <- sprintf("%0.8f", min(df_descStats[which(!is.na(df_descStats[,i])),i])*(1+99*(d==2))) # Min value
        descriptive_stats[i-1,8] <- sprintf("%0.8f", max(df_descStats[which(!is.na(df_descStats[,i])),i])*(1+99*(d==2))) # Max value
        descriptive_stats[i-1,9] <- sprintf("%0.8f", skewness(df_descStats[which(!is.na(df_descStats[,i])),i])) # Skew
        descriptive_stats[i-1,10] <- sprintf("%0.8f", kurtosis(df_descStats[which(!is.na(df_descStats[,i])),i])-3) # Excess kurt
        descriptive_stats[i-1, 11] <- jarque.bera.test(df_descStats[which(!is.na(df_descStats[,i])),i])$statistic # JB test
        descriptive_stats[i-1, 12] <- ur.df(df_descStats[which(!is.na(df_descStats[,i])),i], type = "none", lags = 0)@teststat[1] # DF
        descriptive_stats[i-1, 13] <- ur.df(df_descStats[which(!is.na(df_descStats[,i])),i], type = "none", lags = 1)@teststat[1] # Aug DF
      } else {
        descriptive_stats[i-1,3] <- sprintf("%0.8f", mean(df_descStats[which(!is.na(df_descStats[,i])),i])*(1+99*(d==2))) # Mean
        descriptive_stats[i-1,4] <- sprintf("%0.8f", sd(df_descStats[which(!is.na(df_descStats[,i])),i])*(1+99*(d==2))) # Std. dev.
        descriptive_stats[i-1,5] <- sprintf("%0.8f", min(df_descStats[which(!is.na(df_descStats[,i])),i])*(1+99*(d==2))) # Min value
        descriptive_stats[i-1,6] <- sprintf("%0.8f", max(df_descStats[which(!is.na(df_descStats[,i])),i])*(1+99*(d==2))) # Max value
        descriptive_stats[i-1,7] <- sprintf("%0.8f", skewness(df_descStats[which(!is.na(df_descStats[,i])),i])) # Skew
        descriptive_stats[i-1,8] <- sprintf("%0.8f", kurtosis(df_descStats[which(!is.na(df_descStats[,i])),i])-3) # Excess kurt
        descriptive_stats[i-1,9] <- paste("$", format(round(cor(df_descStats[,i], df_descStats[,2]), 2),
                                                      nsmall = 2), "$", sep = "") # Excess kurt
        descriptive_stats[i-1, 10] <- paste("$", format(round(jb_test$statistic,2), nsmall = 2), "$", sep = "")
        # getSignificanceStars(jb_test$p.value), "$", sep = "") # Jarque-Bera
        descriptive_stats[i-1, 11] <- paste("$", format(round(df_test@teststat[1],2), nsmall = 2), "$", sep = "")
        # getTauSignificanceStars(df_test@teststat[1], df_test@cval), "$", sep = "") # Jarque-Bera
        descriptive_stats[i-1, 12] <- paste("$", format(round(adf_test@teststat[1],2), nsmall = 2), "$", sep = "")
        # getTauSignificanceStars(adf_test@teststat[1], adf_test@cval), "$", sep = "") # Jarque-Bera
        descriptive_stats[i-1, 13] <- paste("$", format(round(lb_test$statistic,2), nsmall = 2), "$", sep = "") # Ljung-box
        # descriptive_stats[i-1, 10] <- ur.df(df_descStats[which(!is.na(df_descStats[,i])),i], type = "none", lags = 0)@teststat[1] # DF
        # descriptive_stats[i-1, 11] <- ur.df(df_descStats[which(!is.na(df_descStats[,i])),i], type = "none", lags = 1)@teststat[1] # Aug DF
      }
    }
    
    # print(descriptive_stats)
    
    ##  Remove NaNs and infs
    ##  descriptive_stats <- descriptive_stats[is.finite(rowSums(descriptive_stats[,5:dim(descriptive_stats)[2]])),]
    if (useStargazer) {
      latexTableText <- capture.output(stargazer(descriptive_stats, summary = FALSE, rownames = FALSE,
                                                 title = mainNames[d], digits = 2, digit.separate = 3))
      if (d==1) {
        latexTableText[grep("begin[{]tabular[}]", latexTableText)] <- paste("\\begin{tabular}{@{\\extracolsep{5pt}} lc",
                paste(rep("S[scientific-notation=true, round-mode=figures, round-precision=3]", 6), collapse=""), "c",
                paste(rep("S[round-mode=places,round-precision=2]", 4), collapse=""), "} ", sep="")
      } else {
        latexTableText[grep("begin[{]tabular[}]", latexTableText)] <- paste("\\begin{tabular}{@{\\extracolsep{5pt}} lc",
                paste(rep("S[round-mode=places,round-precision=2]", 7), collapse=""),
                paste(rep("S[round-mode=places,round-precision=2]", 4), collapse=""), "} ", sep="")
      }
      mainIndex <- grep("names &", latexTableText)
      tableTextCopy <- latexTableText[(mainIndex + 1):length(latexTableText)]
      insertedMainText <- c(paste("\\multicolumn{2}{c}{\\textbf{\\textsc{General information}}} &",
                                  "\\multicolumn{7}{c}{\\textbf{\\textsc{Descriptive statistics}}} &",
                                  "\\multicolumn{4}{c}{\\textbf{\\textsc{Tests}}} \\\\", sep = ""),
                            "\\cmidrule(lr){1-2} \\cmidrule(lr){3-9} \\cmidrule(lr){10-13}",
                            paste("{", paste(latexTableColNames, collapse = "} & {"), "} \\\\", sep = ""), # First titles
                            paste(paste(latexTableUnits[[d]], collapse = " & "), " \\\\", sep = ""))
      latexTableText <- c(latexTableText[1:(mainIndex-1)], insertedMainText, tableTextCopy)
      
      # Sub out some irritating things
      latexTableText <- gsub("\\\\textbackslash", "\\", latexTableText)
      latexTableText <- gsub("Delta", "\\\\Delta", latexTableText)
      latexTableText <- gsub("Delta ln", "Delta \\\\ln", latexTableText)
      # latexTableText <- gsub("\\\\textasteriskcentered ", "*", latexTableText)
      # latexTableText <- gsub("\\$\\\\hat\\{\\\\mkern6mu\\}\\$", "\\^", latexTableText)
      latexTableText <- gsub("\\\\\\$", "$", latexTableText)
      latexTableText <- gsub("\\$-\\$", "-", latexTableText)
      latexTableText <- gsub("\\\\\\_", "_", latexTableText)
      latexTableText <- gsub("\\\\\\{", "{", latexTableText)
      latexTableText <- gsub("\\\\\\}", "}", latexTableText)
      
      # Grep out dollars
      latexTableText <- gsub("\\& \\$", "& ", latexTableText)
      latexTableText <- gsub("\\$ \\&", " &", latexTableText)
      latexTableText <- gsub("\\$ \\\\", " \\\\", latexTableText)
      # latexTableText <- gsub("\\\\rho=6", "$\\\\rho=6$", latexTableText)
      latexTableText <- gsub("\\\\Delta", "$\\\\Delta", latexTableText)
      
      if (addResizeBox) {
        latexTableText[latexTableText == "  \\label{} "] = paste("  \\label{tab:", listNames[d],
                                                                 "} \\begin{adjustbox}{max width=", resizeBoxFactor, "\\textheight}", sep = "")
        latexTableText[latexTableText == "\\end{tabular} "] = "\\end{tabular} \\end{adjustbox}"
      }
      # Add p-values and tau values
      tau_df <- df_test@cval
      tau_adf <- adf_test@cval
      latexTableText[which(latexTableText == "\\hline \\\\[-1.8ex] "  )[3]] = paste("\\hline \\hline \\\\[-1.8ex] \\textit{Critical values:}  & \\multicolumn{",
                                                                                    (dim(descriptive_stats)[2]-1), "}{r}{JB: $\\chi_{2, \\alpha = 10\\%}^2 > ", 
                                                                                    format(round(qchisq(0.90, 2), 2), nsmall = 2), "$", 
                                                                                    "; $\\chi_{2, \\alpha = 5\\%}^2 > ", format(round(qchisq(0.95, 2), 2), nsmall = 2), "$",
                                                                                    "; $\\chi_{2, \\alpha = 1\\%}^2 > ", format(round(qchisq(0.99, 2), 2), nsmall = 2), "$} \\\\ ",
                                                                                    "\\multicolumn{",
                                                                                    (dim(descriptive_stats)[2]), "}{r}{DF, ADF: $\\tau_{\\alpha = 10\\%}<", tau_df[3], "$; ", 
                                                                                    "$\\tau_{\\alpha = 5\\%}<", tau_df[2], "$; $\\tau_{\\alpha = 1\\%}<", tau_df[1],
                                                                                    "$} \\\\ ", 
                                                                                    "\\multicolumn{",
                                                                                    (dim(descriptive_stats)[2]), "}{r}{Ljung-Box: $\\chi_{", lb_test$parameter,", \\alpha = 10\\%}^2 > ", 
                                                                                    format(round(qchisq(0.90, lb_test$parameter), 2), nsmall = 2), "$", 
                                                                                    "; $\\chi_{", lb_test$parameter, ", \\alpha = 5\\%}^2 > ", format(round(qchisq(0.95, lb_test$parameter), 2), nsmall = 2), "$",
                                                                                    "; $\\chi_{", lb_test$parameter, ", \\alpha = 1\\%}^2 > ", format(round(qchisq(0.99, lb_test$parameter), 2), nsmall = 2), "$} \\\\ ",
                                                                                    sep = "")
      latexTableText[grep("begin[{]table[}]", latexTableText)] <- paste("\\begin{landscape}", "\\begin{table}[!htbp] \\centering ", sep = " ")
      latexTableText[grep("end[{]table[}]", latexTableText)] <- paste("\\end{table} "  , "\\end{landscape}", sep = " ")
      
      latexFileName <- paste(outFileDir, listNames[d], ".tex", sep = "")
      print(paste("Writing to: ", latexFileName, sep = ""))
      write(latexTableText, file = latexFileName)
    } else {
      print(paste("Writing to: ", outFileAddress, sep = ""))
      
      # Now write the csv
      write.csv2(descriptive_stats, file=paste(outFileAddress, listNames[d],".csv", sep=""),
                 row.names = F, sep=";", quote=F)
    }
  }
}

constructQrNamesTable <- function(qrModels, horizonVector=c(),
                                  df_outputNames) {
  # First find the appropriate number of series that have been used
  namesRows <- length(qrModels[[1]][[1]]$coefficients)
  for (i in 1:length(qrModels)) {
    for (j in 1:length(qrModels[[i]])) {
      if (length(qrModels[[i]][[j]]$coefficients) > namesRows) {
        namesRows <- length(qrModels[[i]][[j]]$coefficients)
      }
    }  
  }
  # Set up the "names" table
  qrNamesTable <- data.frame(matrix("",namesRows, length(qrModels)),
                             stringsAsFactors=F)
  if (length(horizonVector) > 0) {
    colnames(qrNamesTable) <- paste("Horizon", horizonVector)
  } else {
    colnames(qrNamesTable) <- paste("Horizon", seq(1:dim(qrNamesTable)[2]))
  }
  rownames(qrNamesTable) <- paste("startcoeff", seq(1:(namesRows))-1, "endcoeff", sep="")
  # Then fill out the table
  for (i in 1:length(qrModels)) {
    for (j in 1:length(qrModels[[i]][[1]]$coefficients)) {
      seriesName <- strsplit(names(qrModels[[i]][[1]]$coefficients)[j], split="_lag.")[[1]][1]
      if (seriesName == "(Intercept)") {
        qrNamesTable[j,i] = seriesName
      } else {
        lag <- as.numeric(strsplit(strsplit(names(qrModels[[i]][[1]]$coefficients)[j], split="_lag.")[[1]][2], split="_delta.")[[1]][1])
        delta <- if(as.numeric(strsplit(names(qrModels[[i]][[1]]$coefficients)[j], split="_delta.")[[1]][2]) == 12) "YoY" else "MoM"
        qrNamesTable[j,i] <- paste(df_outputNames[which(df_outputNames[,1] == seriesName),2],
                                   " (lag ", lag, ", ", delta, ")", sep="")
      }
    }
  }
  return(qrNamesTable)
}

constructQrMetricsTable <- function(qrModels, horizonVector=c(),
                                    metrics_decimals=2) {
  qrMetricsTable <- data.frame(matrix("", 3, length(qrModels)),
                               stringsAsFactors=F)
  if (length(horizonVector) > 0) {
    colnames(qrMetricsTable) <- paste("Horizon", horizonVector) 
  } else {
    colnames(qrMetricsTable) <- paste("Horizon", seq(1:dim(qrMetricsTable)[2]))
  }
  rownames(qrMetricsTable) <- c("AIC", "BIC", "R1")
  # Fill out the table
  for (i in 1:length(qrModels)) {
    aic <- c()
    bic <- c()
    r1  <- c()
    for (j in 1:length(qrModels[[i]])) {
      aic <- c(aic, AIC(qrModels[[i]][[j]], k=2))
      bic <- c(bic, AIC(qrModels[[i]][[j]], k=log(length(qrModels[[i]][[j]]$fitted.values))))
      r1 <-  c(r1, pseudo_rsq(qrModels[[i]][[j]]))
    }
    qrMetricsTable[1,i] <- paste("startmetric", format(round(mean(aic), metrics_decimals), nsmall=metrics_decimals),
                                 ", sd. ", format(round(sd(aic), metrics_decimals), nsmall=metrics_decimals), "endmetric", sep="")
    qrMetricsTable[2,i] <- paste("startmetric", format(round(mean(bic), metrics_decimals), nsmall=metrics_decimals),
                                 ", sd. ", format(round(sd(bic), metrics_decimals), nsmall=metrics_decimals), "endmetric", sep="")
    qrMetricsTable[3,i] <- paste("startmetric", format(round(mean(r1), metrics_decimals), nsmall=metrics_decimals),
                                 ", sd. ", format(round(sd(r1), metrics_decimals), nsmall=metrics_decimals), "endmetric", sep="")
  }
  return(qrMetricsTable)
}

createInSampleQrSummaryTable <- function(qrModels,
                                        outFileDir = '/Users/Kristian/Kristian/Dropbox/Apps/ShareLaTeX/Masteroppgave/tables/',
                                        addResizeBox = TRUE, addLandscape = TRUE,
                                        nQrTableCols = 3,
                                        metrics_decimals = 2,
                                        df_outputNames) {
  require('stargazer')
  latexTableText <- ""
  nQrTableRows <- (length(qrModels) - (length(qrModels) %% nQrTableCols))/nQrTableCols
  nQrTableRows <- nQrTableRows + ((length(qrModels) %% nQrTableCols) > 0)
  start <- 1
  stop <- nQrTableCols
  for (i in 1:nQrTableRows) {
    qrModelSubList <- list()
    counter <- 1
    for (j in start:stop) {
      if (j <= length(qrModels)) {
        qrModelSubList[[counter]] <- qrModels[[j]]
        counter <- counter + 1
      }
    }
    qrNamesTable    <- constructQrNamesTable(qrModelSubList, horizonVector=seq(start, stop),
                                              df_outputNames=df_outputNames)
    qrMetricsTable  <- constructQrMetricsTable(qrModelSubList, horizonVector=seq(start, stop),
                                               metrics_decimals=metrics_decimals)
    # Latex merge
    namesText <- capture.output(stargazer(qrNamesTable, summary = FALSE, rownames = TRUE,
                                          title = "In-sample Quantile Regression Results", digits = 2, digit.separate = 3))
    metricsTableText <- capture.output(stargazer(qrMetricsTable, summary = FALSE, rownames = TRUE,
                                                 title = "", digits = 2, digit.separate = 3))
    hlines <- grep("hline", namesText)
    if (i == 1) {
      latexTableText <- namesText[1:(hlines[4]-1)]
      latexTableText[grep("begin\\{tabular\\}", latexTableText)] <- paste("\\begin{tabular}{@{\\extracolsep{5pt}} ",
                                                                          "c", strrep("l", nQrTableCols), "}", sep="")
    } else {
      namesText <- c(paste("\\hline \\hline ", 
                           "\\multicolumn{", (1 + nQrTableCols), "}{c}{} \\\\ ",
                           "\\multicolumn{", (1 + nQrTableCols), "}{c}{} \\\\ ",
                           "\\hline \\hline", sep=""),
                     namesText[(hlines[2]+1):(hlines[4]-1)])
      latexTableText <- c(latexTableText, namesText)
    }
    latexTableText <- c(latexTableText, "\\hline ")
    hlines_metrics <- grep("hline", metricsTableText)
    latexTableText <- c(latexTableText, metricsTableText[(hlines_metrics[3]+1):
                                                         (hlines_metrics[4]-1)])
    
    # Increment
    start <- start + nQrTableCols
    stop <- stop + nQrTableCols
  }
  latexTableText <- c(latexTableText, "\\hline \\hline", "\\end{tabular}",
                      "\\end{table}")
  
  # Sub out irritating things
  latexTableText <- gsub("\\\\\\$", "$", latexTableText)
  latexTableText <- gsub("\\$-\\$", "-", latexTableText)
  
  # Center subheaders with 'horizon'
  subHeader_indices <- grep("Horizon ", latexTableText)
  for (i in 1:length(subHeader_indices)) {
    subHeader <- latexTableText[subHeader_indices[i]]
    subHeader <- strsplit(subHeader, split=" &")[[1]]
    subHeader <- gsub("\\\\", "", subHeader)
    subHeader <- gsub("  ", "", subHeader)
    subHeader_new <- c()
    for (j in 1:length(subHeader)) {
      subHeader[j] <- gsub("Horizon", "\\\\multicolumn{1}{c}{Horizon", subHeader[j])
      if (grepl("Horizon", subHeader[j])) {
        subHeader[j] <- paste(subHeader[j], "}", sep="")
      }
      subHeader_new <- c(subHeader_new, subHeader[j])
    }
    subHeader_new <- paste(subHeader_new, collapse=" & ")
    subHeader_new <- paste(subHeader_new, "\\\\", sep=" ")
    latexTableText[subHeader_indices[i]] <- subHeader_new
  }
  
  # Center numeric/mathematical expressions
  latexTableText <- gsub("startmetric", " \\\\multicolumn{1}{c}{", latexTableText)
  latexTableText <- gsub("endmetric", "} ", latexTableText)
  
  # Change the R1 metric label
  latexTableText <- gsub("R1", "$R^{1}(\\\\tau)$", latexTableText)
  
  # Change the coefficient labels
  latexTableText <- gsub("startcoeff", "$\\\\beta_{", latexTableText)
  latexTableText <- gsub("endcoeff", "}$", latexTableText)
  
  # Add resizing box
  if (addResizeBox) {
    adjustSize <- if (addLandscape) "max width=1.6\\textheight" else "max width=1.00\\textwidth"
    latexTableText[latexTableText == "  \\label{} "] = paste("  \\label{tab:", "in-sample-qr",
                                                             "} \\begin{adjustbox}{", adjustSize, "}", sep = "")
    latexTableText[latexTableText == "\\end{tabular}"] = "\\end{tabular} \\end{adjustbox}"
  }
  
  # Add landscape mode
  if (addLandscape) {
    latexTableText[grep("begin[{]table[}]", latexTableText)] <- paste("\\begin{landscape}", "\\begin{table}[!htbp] \\centering ", sep = " ")
    latexTableText[grep("end[{]table[}]", latexTableText)] <- paste("\\end{table} "  , "\\end{landscape}", sep = " ")
  }
  
  # Write to file
  latexFileName <- paste(outFileDir, "in-sample-qr", ".tex", sep = "")
  write(latexTableText, file = latexFileName)
}

getPValueStars <- function(pvalue, pvalues) {
  pstars <- ""
  for (p in 1:length(pvalues)) {
    if (pvalue < pvalues[p]) {
      if (nchar(pstars) == 0) {
        pstars <- "^{*"
      } else {
        pstars <- paste(pstars, "*", sep="")
      }
    }
  }
  if (nchar(pstars) > 0) {
    pstars <- paste(pstars, "}", sep="")
  }
  return(pstars)
}

createInSampleQrCoefficientsTable <- function(qrModels,
                                              outFileDir = '/Users/Kristian/Kristian/Dropbox/Apps/ShareLaTeX/Masteroppgave/tables/',
                                              tablesPerPage = 2,
                                              decimals_coefficients = 3,
                                              decimals_significances = 3,
                                              decimals_goodness = 3,
                                              pvalues = c(0.10, 0.05, 0.01),
                                              addLandscape = TRUE,
                                              addResizeBox = TRUE,
                                              landscapeHeightMultiplier = 1.00, portraitWidthMultiplier = 1.20,
                                              df_outputNames) {
  require('stargazer')

  nQrTableCols <- length(qrModels[[1]])
  nQrTableRows <- length(qrModels)
  
  nPages <- ceiling(length(qrModels)/tablesPerPage)
  tableCounter <- 1
  
  for (page in 1:nPages) {
    latexTableText <- ""
    for (i in tableCounter:min(tableCounter+1, length(qrModels))) {
      # First find maxCoeffs
      maxCoeffs <- length(qrModels[[i]][[1]]$coefficients)
      for (j in 1:length(qrModels[[i]])) {
        if (length(qrModels[[i]][[j]]$coefficients) > maxCoeffs) {
          maxCoeffs <- length(qrModels[[i]][[j]]$coefficients)
        }
      }
      
      # Construct the table's respective row names ...
      qrModelSubTable <- data.frame(matrix("", (2*maxCoeffs + 3), nQrTableCols),
                                    stringsAsFactors=F)
      index <- 0
      for (j in 1:(2*maxCoeffs)) {
        if (j %% 2 != 0) {
          index <- index + 1
          addName <- ""
        } else {
          addName <- "-1"
        }
        rownames(qrModelSubTable)[j] <- paste(names(qrModels[[i]][[1]]$coefficients)[index], addName, sep ="")
      }
      rownames(qrModelSubTable)[(2*maxCoeffs)+1] <- "& AIC"
      rownames(qrModelSubTable)[(2*maxCoeffs)+2] <- "& BIC"
      rownames(qrModelSubTable)[(2*maxCoeffs)+3] <- "& R1"
      
      # ... and then the respective col names
      for (k in 1:dim(qrModelSubTable)[2]) {
        colnames(qrModelSubTable)[k] <- paste("$\\tau=", format(qrModels[[i]][[k]]$tau, nsmall=2), "$", sep="")
      }
  
      # Fill inn coefficients, significances and p-values
      for (j in 1:nQrTableCols) {
        # print(paste("Page ", page, " - i=", i, ", j=", j, sep=""))
        qrModel <- qrModels[[i]][[j]]
        qrModel_summary <- summary(qrModel, se="boot", bsmethod="pwy", R=1000) # 2017-05-14 Parzen robust standard errors
        for (k in 1:length(qrModel$coefficients)) {
          pstars <- getPValueStars(qrModel_summary$coefficients[k,4], pvalues)
          qrModelSubTable[2*k-1,j] <- paste("$", format(round(as.numeric(qrModel$coefficients[k]), decimals_coefficients),
                                                       nsmall = decimals_coefficients), pstars, "$", sep = "")
          qrModelSubTable[2*k,j]   <- paste("$scriptstyle{(", format(round(qrModel_summary$coefficients[k,3], decimals_significances),
                                                             nsmall = decimals_significances), ")}$", sep = "")
        }
        qrModelSubTable[(2*k+1),j] <- paste("$", format(round(AIC(qrModel, k=2), decimals_goodness), nsmall = decimals_goodness), "$", sep = "")
        qrModelSubTable[(2*k+2),j] <- paste("$", format(round(AIC(qrModel, k=log(length(qrModel$fitted.values))), decimals_goodness), nsmall = decimals_goodness), "$", sep = "")
        qrModelSubTable[(2*k+3),j] <- paste("$", format(round(pseudo_rsq(qrModel), decimals_goodness), nsmall = decimals_goodness), "$", sep = "")
        
        print(paste("Old BIC: ", AIC(qrModels[[i]][[j]], k=0), sep=""))
        print(paste("New BIC @ k=", log(length(qrModels[[i]][[j]]$fitted.values)) ,": ", AIC(qrModels[[i]][[j]], k=log(length(qrModels[[i]][[j]]$fitted.values))), sep=""))
      }
      
      # Generate the LaTeX table
      qrModelSubTableText <- capture.output(stargazer(qrModelSubTable, summary = FALSE, rownames = TRUE, 
                                                      title = paste("Coefficients overview of in-sample quantile regressions across horizons ",
                                                                    "$h \\in \\{1, \\ldots, ", length(qrModels), "\\\\}$ ",
                                                                    "(p. ", page, " of " , nPages, ")", sep="")))
      
      # Sub out irritating stuff...
      qrModelSubTableText <- gsub("\\\\textbackslash ", "\\\\", qrModelSubTableText)
      qrModelSubTableText <- gsub("\\\\textasteriskcentered ", "*", qrModelSubTableText)
      qrModelSubTableText <- gsub("scriptstyle\\\\", "\\\\scriptstyle", qrModelSubTableText)
      qrModelSubTableText <- gsub("\\$\\\\hat\\{\\\\mkern6mu\\}\\$\\\\", "^", qrModelSubTableText)
      qrModelSubTableText <- gsub("\\\\}", "}", qrModelSubTableText)
      qrModelSubTableText <- gsub("\\\\\\$", "$", qrModelSubTableText)
      
      # Fix the header
      qrModelSubTableText[grep("tau", qrModelSubTableText)[1]] <- paste("\\multicolumn{2}{c}{\\textbf{Horizon ", i, "}}",
                                                                          qrModelSubTableText[grep("tau", qrModelSubTableText)[1]], sep="")
      
      # Loop through the row names, rename in accordance with df_outputNames
      startRowLoopIndex <- grep("hline", qrModelSubTableText)[3]+1
      endRowLoopIndex   <- grep("AIC", qrModelSubTableText)[1]-1
      coefficientRows <- qrModelSubTableText[startRowLoopIndex:endRowLoopIndex]
      for (j in 1:length(coefficientRows)) {
        ampersandIndex <- gregexpr(pattern = "&", text = coefficientRows[j], ignore.case = F)[[1]][1]
        rowText <- substr(coefficientRows[j], 1, ampersandIndex - 1)
        if (grepl("-", rowText)) {
          newRowText <- "& "
        } else {
          if (!grepl("Intercept", rowText)) {
            rowText <- gsub("\\\\", "", rowText)
            lag <- as.numeric(strsplit(strsplit(rowText, split="_lag.")[[1]][2], split="_delta.")[[1]][1])
            delta <- if(as.numeric(strsplit(rowText, "_delta.")[[1]][2]) == 12) "YoY" else "MoM"
            rowSeriesName <- strsplit(rowText, split="_lag.")[[1]][1]
            newRowText <- paste(df_outputNames[which(df_outputNames[,1] == rowSeriesName), 2],
                                " (lag ", lag, ", ", delta, ")", sep="")
            newRowText <- gsub("#", "\\\\#", newRowText)
            newRowText <- gsub("%", "\\\\%", newRowText)
            newRowText <- paste("\\multirow{2}{*}{$\\beta_{", (j-1)/2, "}$} & \\multirow{2}{*}{\\begin{minipage}{2.5in}", newRowText, "\\end{minipage}}", sep="")
          } else {
            newRowText <- paste("\\multirow{2}{*}{$\\beta_{", 0, "}$} & \\multirow{2}{*}{(Intercept)}", sep="")
          }
        }
        rowText <- paste(newRowText, " ", substr(coefficientRows[j], ampersandIndex, nchar(coefficientRows[j])))
        qrModelSubTableText[startRowLoopIndex + j - 1] <- rowText
      }
      
      # Construct the full LaTeX table
      if (i == tableCounter) {
        latexTableText <- c(qrModelSubTableText[1:(grep("AIC", qrModelSubTableText)[1]-1)], "\\hline \\\\[-1.8ex]",
                            qrModelSubTableText[(grep("AIC", qrModelSubTableText)[1]):(grep("R1", qrModelSubTableText)[1])],
                            "\\hline")
      } else {
        latexTableText <- c(latexTableText, "\\hline", qrModelSubTableText[grep("hline", qrModelSubTableText)[2]:(grep("AIC", qrModelSubTableText)[1]-1)],
                            "\\hline \\\\[-1.8ex]", qrModelSubTableText[(grep("AIC", qrModelSubTableText)[1]):(grep("R1", qrModelSubTableText)[1])],
                            "\\hline")
      }
      
      # Add some spacing
      if (i < min(tableCounter+1, length(qrModels))) {
        latexTableText <- c(latexTableText, "\\\\")
      } else {
        latexTableText <- c(latexTableText, paste("\\\\[-1.8ex] & \\textit{Note:} & \\multicolumn{",
                            nQrTableCols, "}{r}{$^{*}$p$<$", pvalues[1], "; $^{**}$p$<$",
                            pvalues[2], "; $^{***}$p$<$", pvalues[3], "}", sep=""),
                            "\\\\")
      }
      
    }
    latexTableText <- c(latexTableText, "\\end{tabular}", "\\end{table}")
    
    # Change the R1 metric label
    latexTableText <- gsub("R1", "$R^{1}(\\\\tau)$", latexTableText)
    
    # Change the column centering
    latexTableText[grep("begin\\{tabular\\}", latexTableText)] <- paste("\\begin{tabular}{@{\\extracolsep{5pt}} ",
                                                              "cl", paste(rep("c", nQrTableCols), collapse=""), "}", sep="")
    
    # Add resizing box
    if (addResizeBox) {
      if (addLandscape) {
        adjustSize <- paste("max width=", landscapeHeightMultiplier, "\\textheight", sep="")
      } else {
        adjustSize <- paste("max width=", portraitWidthMultiplier, "\\textwidth", sep="")
      }
      latexTableText[latexTableText == "  \\label{} "] = paste("  \\label{tab:", "in-sample-qr-coefficients",
                                                               "} \\begin{adjustbox}{", adjustSize, "}", sep = "")
      latexTableText[latexTableText == "\\end{tabular}"] = "\\end{tabular} \\end{adjustbox}"
    }
    
    # Add landscape mode
    if (addLandscape) {
      latexTableText[grep("begin[{]table[}]", latexTableText)] <- paste("\\begin{landscape}", "\\begin{table}[!htbp] \\centering ", sep = " ")
      latexTableText[grep("end[{]table[}]", latexTableText)] <- paste("\\end{table} "  , "\\end{landscape}", sep = " ")
    }
    
    latexFileName <- paste(outFileDir, "in-sample-qr-coefficients-page-", page, ".tex", sep = "")
    write(latexTableText, file = latexFileName)
    
    # Page count
    tableCounter <- tableCounter + tablesPerPage
  }
}

qrCoverageRate <- function(qrModel, tau) {
  n   <- length(qrModel$fitted.values)
  n_1 <- sum(qrModel$fitted.values > qrModel$fitted.values + qrModel$residuals)
  pi_obs <- n_1/n
  return(pi_obs)
}

qrKupiec <- function(qrModel, tau) {
  pi_exp   <- tau
  n   <- length(qrModel$fitted.values)
  n_1 <- sum(qrModel$fitted.values > qrModel$fitted.values + qrModel$residuals)
  n_0 <- n - n_1
  pi_obs <- n_1/n
  kupiecLL <- -2*log((pi_exp^n_1)*((1-pi_exp)^n_0)/((pi_obs^n_1)*((1-pi_obs)^n_0)))
  return(kupiecLL)
}

qrChristoffersen <- function(qrModel) {
  fit <- qrModel$fitted.values > qrModel$fitted.values + qrModel$residuals
  tau <- qrModel$tau
  n   <- length(fit)
  n_00 <- sum(as.logical(fit[1:(n-1)] == F) * as.logical(fit[2:n] == F))
  n_01 <- sum(as.logical(fit[1:(n-1)] == F) * as.logical(fit[2:n]))
  n_10 <- sum(as.logical(fit[1:(n-1)]) * as.logical(fit[2:n] == F))
  n_11 <- sum(as.logical(fit[1:(n-1)]) * as.logical(fit[2:n]))
  
  n_1 <- sum(qrModel$fitted.values > qrModel$fitted.values + qrModel$residuals)
  n_0 <- n - n_1
    
  pi_obs <- sum(fit)/n
  pi_exp <- tau
  pi_01  <- n_01/(n_00+n_01)
  pi_11  <- n_11/(n_10+n_11)
  if ((pi_11 > 0) && (pi_01 < 1)) {
    # print(paste("n_00:", n_00))
    # print(paste("n_01:", n_01))
    # print(paste("n_10:", n_10))
    # print(paste("n_11:", n_11))
    # print(paste("pi_01:", pi_01))
    # print(paste("pi_11:", pi_11))
    # print(paste("pi_exp:", pi_exp))
    christoffersenLL <- -2*(n_0*log(1-pi_exp) + n_1*log(pi_exp)
                            - n_00*log(1-pi_01) - n_01*log(pi_01)
                            - n_10*log(1-pi_11) - n_11*log(pi_11))
  } else if ((pi_11 == 0) && (pi_01 < 1)) {
    christoffersenLL <- -2*(n_0*log(1-pi_exp) + n_1*log(pi_exp)
                            - n_00*log(1-pi_01) - n_01*log(pi_01))
  } else if ((pi_11 > 0) && (pi_01 == 1)) {
    christoffersenLL <- -2*(n_0*log(1-pi_exp) + n_1*log(pi_exp)
                            - n_10*log(1-pi_11) - n_11*log(pi_11))
  } else {
    christoffersenLL <- "N/A"
  }
  return(christoffersenLL)
}

getQrChristoffersenStatisticString <- function(qrModel, chi_sq_significance_levels, metrics_decimals) {
  fit <- qrModel$fitted.values > qrModel$fitted.values + qrModel$residuals
  tau <- qrModel$tau
  n   <- length(fit)
  n_00 <- sum(as.logical(fit[1:(n-1)] == F) * as.logical(fit[2:n] == F))
  n_01 <- sum(as.logical(fit[1:(n-1)] == F) * as.logical(fit[2:n]))
  n_10 <- sum(as.logical(fit[1:(n-1)]) * as.logical(fit[2:n] == F))
  n_11 <- sum(as.logical(fit[1:(n-1)]) * as.logical(fit[2:n]))
  
  n_1 <- sum(qrModel$fitted.values > qrModel$fitted.values + qrModel$residuals)
  n_0 <- n - n_1
  
  pi_obs <- sum(fit)/n
  pi_exp <- tau
  pi_01  <- n_01/(n_00+n_01)
  pi_11  <- n_11/(n_10+n_11)
  if ((pi_11 > 0) && (pi_01 < 1)) {
    # print(paste("n_00:", n_00))
    # print(paste("n_01:", n_01))
    # print(paste("n_10:", n_10))
    # print(paste("n_11:", n_11))
    # print(paste("pi_01:", pi_01))
    # print(paste("pi_11:", pi_11))
    # print(paste("pi_exp:", pi_exp))
    christoffersenLL <- -2*(n_0*log(1-pi_exp) + n_1*log(pi_exp)
                            - n_00*log(1-pi_01) - n_01*log(pi_01)
                            - n_10*log(1-pi_11) - n_11*log(pi_11))
    altFlag <- ""
  } else if ((pi_11 == 0) && (pi_01 < 1)) {
    christoffersenLL <- -2*(n_0*log(1-pi_exp) + n_1*log(pi_exp)
                            - n_00*log(1-pi_01) - n_01*log(pi_01))
    altFlag <- "\\phantom{}^{\\dag}"
    # print(christoffersenLL)
  } else if ((pi_11 > 0) && (pi_01 == 1)) {
    christoffersenLL <- -2*(n_0*log(1-pi_exp) + n_1*log(pi_exp)
                            - n_10*log(1-pi_11) - n_11*log(pi_11))
    altFlag <- "\\phantom{}^{\\ddag}"
  } else {
    christoffersenLL <- "N/A"
  }
  # print(paste("Found likelihood: ", christoffersenLL, " ", altFlag, sep=""))
  
  qrChristoffersenStatisticString <- paste("$", format(round(christoffersenLL, metrics_decimals), nsmall=metrics_decimals),
                                           getChiSqStars(christoffersenLL, chi_sq_significance_levels, df=2), 
                                           altFlag, "$", sep="")
  return(qrChristoffersenStatisticString)
}

getChiSqStars <- function(testStatistic, chi_sq_significance_levels, df=1) {
  chiSqStars <- ""
  
  for (i in 1:length(chi_sq_significance_levels)) {
    sign_level <- 1-chi_sq_significance_levels[i]
    if (qchisq(sign_level, df=df) < testStatistic) {
      chiSqStars <- paste("^{", paste(rep("*", i), collapse=""), "}", sep="")
    }
  }
  return(chiSqStars)
}

qrKernelConsistentSpecTest <- function(qrModel) {
  require(np)
  require(quantreg)
  testmodel <- rq(data.frame(cbind(qrModel$y), qrModel$x[,2:dim(qrModel$x)[2]]), tau=qrModel$tau, model=TRUE)
  testres <- npqcmstest(model=testmodel, xdat=qrModel$x[,2:dim(qrModel$x)[2]], ydat=qrModel$y, tau=qrModel$tau, distribution="bootstrap")
  return(testres)
}

getSignificanceStars <- function (pvalue, significance_levels) {
  stars <- ""
  for (i in 1:length(significance_levels)) {
    sign_level <- 1 - significance_levels[i]
    if ((1-pvalue) >= sign_level) {
      stars <- paste("^{", paste(rep("*", i), collapse=""), "}", sep="")
    }
  }
  return(stars)
}


createInSampleSpecificationTable <- function(qrModels, Tau=c(0.10, 0.25, 0.50, 0.75, 0.90),
                                             outFileDir = '/Users/Kristian/Kristian/Dropbox/Apps/ShareLaTeX/Masteroppgave/tables/',
                                             addResizeBox = TRUE, addLandscape = FALSE,
                                             landscapeHeightMultiplier = 0.28, portraitWidthMultiplier = 1.20,
                                             nQrTableCols = 2, metrics_decimals = 3,
                                             chi_sq_significance_levels = c(0.10, 0.05, 0.01), qreset_orders=c(2,3,4),
                                             useNpqcmstest = TRUE, addDashedSeparatorLines = TRUE,
                                             subTableHSpace = 6.0, subTableMiniPageWidth=0.45, subTableCaptionFontSize="large",
                                             df_outputNames) {
  require('stargazer')
  
  nQrTableRows <- ceiling(length(qrModels)/nQrTableCols)
  rowCounter <- 1
  
  for (i in 1:length(qrModels)) {
    horizonModelSpecTable <- data.frame(matrix("", 3 + length(qreset_orders) + 3*useNpqcmstest, length(qrModels[[i]])+1), stringsAsFactors=F)
    colnames(horizonModelSpecTable) <- c("Horizon", paste("$\\tau=", format(Tau, nsmall=2), "$", sep=""))
    horizonModelSpecTable[1:3,1] <- c("Coverage rate", "Kupiec test", "Christoffersen test")
    
    # QRESET test
    if (length(qreset_orders) > 0) {
      for (j in 1:length(qreset_orders)) {
        horizonModelSpecTable[3+j,1] <- paste("QRESET(", qreset_orders[j], ")", sep="")
      }
    }
    
    # npqcms test
    if (useNpqcmstest) {
      horizonModelSpecTable[3+length(qreset_orders)+1,1] <- paste("Kernel test -- $I_n$")
      horizonModelSpecTable[3+length(qreset_orders)+2,1] <- paste("Kernel test -- $J_n$")
      horizonModelSpecTable[3+length(qreset_orders)+3,1] <- paste("$\\quad$ ($p$\\textit{-value of} $J_n$)")
    }
    
    for (j in 1:length(qrModels[[i]])) {
      tau <- Tau[j]
      qrModel <- qrModels[[i]][[j]]
      qrKupiecStatistic <- qrKupiec(qrModel, tau)
      # qrChristoffersenStatistic <- qrChristoffersen(qrModel, tau)
      
      # Fill out the table...
      horizonModelSpecTable[1,1+j] <- format(round(qrCoverageRate(qrModel, tau), metrics_decimals), nsmall = metrics_decimals)
      horizonModelSpecTable[2,1+j] <- paste("$", format(round(qrKupiecStatistic, metrics_decimals), nsmall=metrics_decimals), getChiSqStars(qrKupiecStatistic, chi_sq_significance_levels, df=1), "$", sep="") 
      horizonModelSpecTable[3,1+j] <- getQrChristoffersenStatisticString(qrModel, chi_sq_significance_levels, metrics_decimals)
      
      if (length(qreset_orders) > 0) {
        for (k in 1:length(qreset_orders)) {
          horizonModelSpecTable[3+k,1+j] <- qreset_to_string(rq_model=qrModel, order=qreset_orders[k], R=1000,
                                                             chi_sq_significance_levels=chi_sq_significance_levels,
                                                             n_decimals=metrics_decimals)
        }
      }
      
      if (useNpqcmstest) {
        testres <- qrKernelConsistentSpecTest(qrModel)
        horizonModelSpecTable[3+length(qreset_orders)+1,1+j] <- paste("$", format(round(testres$In, metrics_decimals), nsmall=metrics_decimals), "$", sep="")
        horizonModelSpecTable[3+length(qreset_orders)+2,1+j] <- paste("$", format(round(testres$Jn, metrics_decimals), nsmall=metrics_decimals), getSignificanceStars(testres$P, chi_sq_significance_levels), "$", sep="")
        horizonModelSpecTable[3+length(qreset_orders)+3,1+j] <- paste("$\\quad$ \\textit{(", format(round(testres$P, metrics_decimals), nsmall=metrics_decimals), ")}", sep="")
      }
      
      # if (!is.nan(qrChristoffersenStatistic)) {
      #   horizonModelSpecTable[3,1+j] <- paste("$", format(round(qrChristoffersenStatistic, metrics_decimals), nsmall=metrics_decimals), getChiSqStars(qrChristoffersenStatistic, chi_sq_significance_levels, df=2), "$", sep="") 
      # } else {
      #   horizonModelSpecTable[3,1+j] <- "N/A"
      # }
      
    }
    horizonModelSpecTableText <- capture.output(stargazer(horizonModelSpecTable, summary = FALSE, rownames = FALSE, 
                           title = paste("Specification test results from the in-sample quantile regressions across horizons ",
                                         "$k \\in \\{1, \\ldots ", length(qrModels), "\\\\}$", sep="")))
    # Sub out irritating stuff...
    horizonModelSpecTableText <- gsub("\\\\textbackslash ", "\\\\", horizonModelSpecTableText)
    horizonModelSpecTableText <- gsub("\\\\textasteriskcentered ", "*", horizonModelSpecTableText)
    horizonModelSpecTableText <- gsub("scriptstyle\\\\", "\\\\scriptstyle", horizonModelSpecTableText)
    horizonModelSpecTableText <- gsub("\\$\\\\hat\\{\\\\mkern6mu\\}\\$\\\\", "^", horizonModelSpecTableText)
    horizonModelSpecTableText <- gsub("\\\\}", "}", horizonModelSpecTableText)
    horizonModelSpecTableText <- gsub("\\\\\\$", "$", horizonModelSpecTableText)
    horizonModelSpecTableText <- gsub("textit\\\\\\{", "textit{", horizonModelSpecTableText)
    
    # Phantoms for math
    horizonModelSpecTableText <- gsub("phantom\\\\\\{", "phantom{", horizonModelSpecTableText)
    # Subscripts for math
    horizonModelSpecTableText <- gsub("\\\\\\_", "_", horizonModelSpecTableText)
    
    # Add dashed separator lines
    if (addDashedSeparatorLines) {
      for (l in grep("Christoffersen", horizonModelSpecTableText)) {
        cat("Grepping Christoffersen...")
        flush.console()
        horizonModelSpecTableText[l] <- paste(horizonModelSpecTableText[l], " \\hdashline", sep="")
      }
      if (length(qreset_orders)) {
        for (l in grep(paste("QRESET\\(", qreset_orders[length(qreset_orders)], sep=""), horizonModelSpecTableText)) {
          cat("Grepping qreset...")
          flush.console()
          horizonModelSpecTableText[l] <- paste(horizonModelSpecTableText[l], " \\hdashline", sep="")
        }
      }
    }
    
    # Add horizon title
    # horizonModelSpecTableText[grep("\\hline", horizonModelSpecTableText)]
    horizonModelSpecTableText <- gsub("Horizon", paste("\\\\multicolumn{1}{c}{\\\\textbf{Horizon ", i, "}}", sep=""), horizonModelSpecTableText)
    if (i == 1) {
      inSampleSpecTableText <- horizonModelSpecTableText[4:(grep("hline", horizonModelSpecTableText)[4])]
      inSampleSpecTableText[grep("begin\\{tabular\\}", inSampleSpecTableText)] <- paste("\\begin{tabular}{@{\\extracolsep{5pt}} ",
                                                                          "l", strrep("c", length(qrModels[[i]])), "}", sep="")
      if ((addLandscape) && (nQrTableCols > 1)) {
        inSampleSpecTableText <- c(inSampleSpecTableText[1:grep("begin\\{tabular\\}", inSampleSpecTableText)-1],
                                   paste("\\begin{minipage}[T]{", subTableMiniPageWidth, "\\textheight}", sep=""),
                                   "\\begin{subtable}{0.45\\textheight}", "\\centering", paste("\\caption{Horizons ", i, " --", nQrTableRows, "}", sep=""),
                                   inSampleSpecTableText[grep("begin\\{tabular\\}", inSampleSpecTableText):length(inSampleSpecTableText)])
      }
    } else if ((i %% nQrTableRows == 0) && (i < length(qrModels)) && addLandscape) { # End column
      inSampleSpecTableText <- c(inSampleSpecTableText, "\\\\ \\hline", horizonModelSpecTableText[9:(length(horizonModelSpecTableText)-2)],
                                 "\\end{tabular}", "\\end{subtable}", "\\end{minipage}", paste("\\hspace{", subTableHSpace, "cm}", sep="")) # End of current subtable, begin new...
    } else if ((i %% nQrTableRows == 1) && addLandscape) {
      horizonModelSpecTableText[grep("begin\\{tabular\\}", horizonModelSpecTableText)] <- paste("\\begin{tabular}{@{\\extracolsep{5pt}} ",
                                                                                        "l", strrep("c", length(qrModels[[i]])), "}", sep="")
      inSampleSpecTableText <- c(inSampleSpecTableText, paste("\\begin{minipage}[T]{", subTableMiniPageWidth, "\\textheight}", sep=""),
                                 "\\begin{subtable}{0.45\\textheight}", "\\centering",
                                 paste("\\caption{Horizons ", i, " --", min(i+nQrTableRows-1, length(qrModels)), "}", sep=""),
                                 horizonModelSpecTableText[grep("begin\\{tabular\\}", horizonModelSpecTableText):(grep("hline", horizonModelSpecTableText)[4])])
    } else if (i < length(qrModels)) {
      inSampleSpecTableText <- c(inSampleSpecTableText, "\\\\ \\hline", horizonModelSpecTableText[9:(grep("hline", horizonModelSpecTableText)[4])])
    } else if (i == length(qrModels)) {
      inSampleSpecTableText <- c(inSampleSpecTableText, "\\\\ \\hline", horizonModelSpecTableText[9:(length(horizonModelSpecTableText)-2)])
      inSampleSpecTableText <- c(inSampleSpecTableText, paste("\\textit{Note:} & \\multicolumn{",
                                                              length(qrModels[[i]]), "}{r}{$^{*}$p$<$", chi_sq_significance_levels[1], "; $^{**}$p$<$",
                                                              chi_sq_significance_levels[2], "; $^{***}$p$<$", chi_sq_significance_levels[3], "}", sep=""), "\\\\")
      # Daggers for Christoffersen
      inSampleSpecTableText <- c(inSampleSpecTableText, paste("\\multicolumn{",
                                                              length(qrModels[[i]])+1, "}{r}{$^{\\dag}\\pi_{11}=0;\\;^{\\ddag}\\pi_{01}=1$}", sep=""), "\\\\")
      # Chi squared, 1 DF
      inSampleSpecTableText <- c(inSampleSpecTableText, paste("\\textit{Critical values:} & \\multicolumn{", length(qrModels[[i]]), "}{r}{",
                                                              "$\\chi_{1, \\alpha=", format(100*chi_sq_significance_levels[1], nsmall=0), "\\%}^2 > ",
                                                              format(round(qchisq(1-chi_sq_significance_levels[1], df=1), 2), nsmall=2), "$",
                                                              "; $\\chi_{1, \\alpha=", format(100*chi_sq_significance_levels[2], nsmall=0), "\\%}^2 > ",
                                                              format(round(qchisq(1-chi_sq_significance_levels[2], df=1), 2), nsmall=2), "$",
                                                              "; $\\chi_{1, \\alpha=", format(100*chi_sq_significance_levels[3], nsmall=0), "\\%}^2 > ",
                                                              format(round(qchisq(1-chi_sq_significance_levels[3], df=1), 2), nsmall=2), "$} \\\\",
                                                              sep=""))
      # Chi squared, 2 DF
      inSampleSpecTableText <- c(inSampleSpecTableText, paste(" & \\multicolumn{", length(qrModels[[i]]), "}{r}{",
                                                              "$\\chi_{2, \\alpha=", format(100*chi_sq_significance_levels[1], nsmall=0), "\\%}^2 > ",
                                                              format(round(qchisq(1-chi_sq_significance_levels[1], df=2), 2), nsmall=2), "$",
                                                              "; $\\chi_{2, \\alpha=", format(100*chi_sq_significance_levels[2], nsmall=0), "\\%}^2 > ",
                                                              format(round(qchisq(1-chi_sq_significance_levels[2], df=2), 2), nsmall=2), "$",
                                                              "; $\\chi_{2, \\alpha=", format(100*chi_sq_significance_levels[3], nsmall=0), "\\%}^2 > ",
                                                              format(round(qchisq(1-chi_sq_significance_levels[3], df=2), 2), nsmall=2), "$} \\\\",
                                                              sep=""))
      # Chi squared, 3 DF
      inSampleSpecTableText <- c(inSampleSpecTableText, paste(" & \\multicolumn{", length(qrModels[[i]]), "}{r}{",
                                                              "$\\chi_{3, \\alpha=", format(100*chi_sq_significance_levels[1], nsmall=0), "\\%}^2 > ",
                                                              format(round(qchisq(1-chi_sq_significance_levels[1], df=3), 2), nsmall=2), "$",
                                                              "; $\\chi_{3, \\alpha=", format(100*chi_sq_significance_levels[2], nsmall=0), "\\%}^2 > ",
                                                              format(round(qchisq(1-chi_sq_significance_levels[2], df=3), 2), nsmall=2), "$",
                                                              "; $\\chi_{3, \\alpha=", format(100*chi_sq_significance_levels[3], nsmall=0), "\\%}^2 > ",
                                                              format(round(qchisq(1-chi_sq_significance_levels[3], df=3), 2), nsmall=2), "$} \\\\",
                                                              sep=""))
      if ((addLandscape) && (nQrTableCols > 1)) {
        inSampleSpecTableText <- c(inSampleSpecTableText, "\\end{tabular}", "\\end{subtable}", "\\end{minipage}","\\end{table}")
      } else {
        inSampleSpecTableText <- c(inSampleSpecTableText, "\\end{tabular}", "\\end{table}")
      }
    }
  }
  
  # Add resizing box
  if (addResizeBox) {
    if (addLandscape) {
      adjustSize <- paste("width=", landscapeHeightMultiplier, "\\textheight", sep="")
    } else {
      adjustSize <- paste("width=", portraitWidthMultiplier, "\\textwidth", sep="")
    }
    if (addLandscape && (nQrTableCols > 1)) {
      resize_box_indices <- grep("begin\\{minipage", inSampleSpecTableText)
      for (p in 1:(length(resize_box_indices))) {
        l <- resize_box_indices[p] + p - 1
        # print(paste(resize_box_indices, collapse=", "))
        # print(paste("l-1=", (l-1), sep=""))
        inSampleSpecTableText <- c(inSampleSpecTableText[1:(l-1)], paste("\\adjustbox{", adjustSize, ", valign=t}{", sep = ""),
                                   inSampleSpecTableText[l:length(inSampleSpecTableText)])
      }
      resize_box_indices <- grep("end\\{minipage", inSampleSpecTableText)
      for (p in 1:(length(resize_box_indices))) {
        l <- resize_box_indices[p] + p - 1
        inSampleSpecTableText <- c(inSampleSpecTableText[1:l], "}",
                                   inSampleSpecTableText[(l+1):length(inSampleSpecTableText)])
      }
    } else {
      inSampleSpecTableText[inSampleSpecTableText == "  \\label{} "] = paste("  \\label{tab:", "in-sample-qr-coefficients",
                                                               "} \\begin{adjustbox}{", adjustSize, "}", sep = "")
      inSampleSpecTableText[inSampleSpecTableText == "\\end{tabular}"] = "\\end{tabular} \\end{adjustbox}"
    }
  }
  
  # Add landscaping
  if (addLandscape) {
    inSampleSpecTableText[grep("begin[{]table[}]", inSampleSpecTableText)] <- paste("\\begin{landscape} \\centering", "\\begin{table}[!htbp] \\captionsetup[sub]{font=",
                                                                                subTableCaptionFontSize, "} \\centering ", sep = " ")
    inSampleSpecTableText[grep("end[{]table[}]", inSampleSpecTableText)] <- paste("\\end{table} "  , "\\end{landscape}", sep = " ")
  }
  
  # Write to file
  latexFileName <- paste(outFileDir, "in-sample-specification-tests", ".tex", sep = "")
  write(inSampleSpecTableText, file = latexFileName)
   
}

































